'use strict';

const AWS = require('aws-sdk');
const sns = new AWS.SNS();

// Environment variable for email notification
const emailAddress = process.env.EMAIL_ADDRESS;
// Environment variable for action  type. Valid options are EMAIL_ONLY, SUPPORT_CASE_ONLY, and SUPPORT_CASE_AND_EMAIL.
const actionType = process.env.ACTION_TYPE;
// Environment variable for SNS Topic ARN
const topicArn = process.env.TOPIC_ARN;
//  Environment variable for Cost Explorer service endpoint
const costexplorerEndpoint = process.env.COST_EXPLORER_ENDPOINT;
//  Environment variable for Cost Explorer service endpoint
const supportEndpoint = process.env.SUPPORT_ENDPOINT;

// Not all the regions have Cost Explorer service endpoints
const ceSplitArray = costexplorerEndpoint.split(".");
const ceEp = new AWS.Endpoint(costexplorerEndpoint);
const costexplorer = new AWS.CostExplorer({endpoint:ceEp,region:ceSplitArray[1]});

// Not all the regions have Support service endpoints
const supportSplitArray = supportEndpoint.split(".");
const supportEp = new AWS.Endpoint(supportEndpoint);
const support = new AWS.Support({endpoint:supportEp,region:supportSplitArray[1]});

const riRequestSubject = "Reserved Instances Migration";
const spRequestSubject = "Savings Plans Migration";
const spRequestBody = "Please migrate the following Savings Plans from the source account to the destination account.";
const riRequestBody = "Please migrate the following Reserved Instances from the source account to the destination account.";

// Return a two digits string for an one or two digits number
function formatToTwoDigits(number) {
  var twoDigits;
  if (number < 10) {
    twoDigits = '0' + number;
  } else {
    twoDigits = '' + number;
  }
  
  return twoDigits;
}

// Check if the source account has unused RI. If yes, send email notification 
// or create support case based on ACTION_TYPE
// Params
//   startDate: String with 'YYYY-MM-DD' format
//   endDate: String with 'YYYY-MM-DD' format
//   sourceAccountId: the AWS account to migrate RI from
//   destinationAccountId: the AWS account to migrate RI to
function processReservationUtilization(startDate,endDate,
            sourceAccountId,destinationAccountId) {
  
  console.log("Enter processReservationUtilization");
  
  //array to hold all the unused RI lease IDs from the source account
  var riLeaseIds = []; 
  
  const now = new Date();

  var params = {
    TimePeriod: {
      End: endDate, 
      Start: startDate
    },
    Filter: { 
      "Dimensions": {
          "Key": "LINKED_ACCOUNT",
          "Values": [
              sourceAccountId
          ]
      }
    },
    GroupBy: [
      {
        Key: 'SUBSCRIPTION_ID',
        Type: "DIMENSION"
      }
    ]
  };
  
	var handleError = function(response) {
		console.log(response.err, response.err.stack); // an error occurred
		throw response.err;
	};

  costexplorer.getReservationUtilization(params).on('success', function handlePage(response) {
    console.log("GetReservationUtilization response \n" + JSON.stringify(response.data)); 
    
    var utilizationsByTime;
    var group;
    
    // Find unused RI in the results
    for (utilizationsByTime of response.data.UtilizationsByTime) {
      for (group of utilizationsByTime.Groups) {
        const endDateTime = new Date(group.Attributes.endDateTime);
        if (endDateTime.getTime() > now.getTime()) {
          riLeaseIds.push(group.Attributes.leaseId);
        }        
      }
    }
    
		// Handle response pagination
		if (response.hasNextPage()) {
			response.nextPage().on('success', handlePage).on('error', handleError).send();
		} else if (riLeaseIds.length > 0) {
      var requestBody = riRequestBody + "\nSource Account: " + sourceAccountId + "\nDestination Account: " + destinationAccountId + "\nRI Lease IDs: " + JSON.stringify(riLeaseIds);
      callAction(emailAddress,riRequestSubject,requestBody);
    }      
  }).on('error', handleError).send();
}

// Check if the source account has unused SP. If yes, send email notification 
// or create support case based on ACTION_TYPE
// Params
//   startDate: String with 'YYYY-MM-DD' format
//   endDate: String with 'YYYY-MM-DD' format
//   sourceAccountId: the AWS account to migrate SP from
//   destinationAccountId: the AWS account to migrate SP to
function processSavingsPlansUtilization(startDate,endDate,
        sourceAccountId,destinationAccountId) {
 
  console.log("Enter processSavingsPlansUtilization");
  
  var savingsPlansArns = []; //array to hold all the unused SP ARNs from the source account
  
  const now = new Date();  
  
  var params = {
    TimePeriod: { 
      End: endDate, 
      Start: startDate	
    },
    Filter: { 
      "Dimensions": {
          "Key": "LINKED_ACCOUNT",
          "Values": [
            sourceAccountId
          ]
      }
    },
  };

	var handleError = function(response) {
		console.log(response.err, response.err.stack); // an error occurred
		throw response.err;
	};

  costexplorer.getSavingsPlansUtilizationDetails(params).on('success', function handlePage(response) {
    console.log("GetSavingsPlansUtilizationDetails response \n" + JSON.stringify(response.data));  

    var utilizationDetail;
    
    // Find unused SP in the results
    for (utilizationDetail of response.data.SavingsPlansUtilizationDetails) {
      const endDateTime = new Date(utilizationDetail.Attributes.EndDateTime);
      if (endDateTime.getTime() > now.getTime()) {
        savingsPlansArns.push(utilizationDetail.SavingsPlanArn);
      }
    }

		// Handle response pagination
		if (response.hasNextPage()) {
			response.nextPage().on('success', handlePage).on('error', handleError).send();
		} else if (savingsPlansArns.length > 0) {
      var requestBody = spRequestBody + "\nSource Account: " + sourceAccountId + "\nDestination Account: " + destinationAccountId + "\nSavings Plans Arns: " + JSON.stringify(savingsPlansArns);
      callAction(emailAddress,spRequestSubject,requestBody);
    }
  }).on('error', handleError).send();
}

// Call actions based ACTION_TYPE
function callAction(emailAddress,subject,message) {
  if (actionType == 'EMAIL_ONLY') {
    sendEmail(subject,message);
  } else if (actionType == 'SUPPORT_CASE_ONLY') {
    createSupportCase(emailAddress,subject,message,false);
  } else if (actionType == 'SUPPORT_CASE_AND_EMAIL') {
    createSupportCase(emailAddress,subject,message,true);
  }
}


// Create an AWS Support Case
// Params:
//   emailAddress: email address for the support case
//   caseSubject: the subject of the support case
//   caseBody: the body of the support case
function createSupportCase(emailAddresses, caseSubject, caseBody, notify) {
  
  console.log("Enter createSupportCase");
  
  var params = {
    communicationBody: caseBody, 
    subject: caseSubject, 
    categoryCode: 'reserved-instances',
    ccEmailAddresses: [
      emailAddresses
    ],
    issueType: 'customer-service',
    serviceCode: 'billing',
    severityCode: 'low'
  };
  
  support.createCase(params, function(err, data) {
    console.log("CreateCase response...");  
    
    if (err) {
      console.log(err, err.stack); // an error occurred
      throw err;
    } else {
      console.log(data);           // successful response
      if (notify) {
        sendEmailWithCaseDisplayId(data.caseId, caseSubject,caseBody);
      }
    }
  });
}

// Get support case display ID and send a message 
// Params:
//   caseId: support case ID
//   subject: subject of the message
//   message: body of the message
function sendEmailWithCaseDisplayId(caseId, subject, message) {
  
  console.log("Enter sendEmailWithCaseDisplayId");
  
  var params = {
    caseIdList: [
      caseId
    ]
  };
  
  support.describeCases(params, function(err, data) {
    console.log("DescribeCases response...");  
    
    if (err) {
      console.log(err, err.stack); // an error occurred
      throw err;
    } else {
      console.log(data);           // successful response
      
      if (data.cases.length > 0) {
        sendEmail("[Case " + data.cases[0].displayId + "] " + subject,message);
      }
    }
  });
}

// Send a message to the SNS topic for email notification
// Params:
//   subject: subject of the message
//   message: body of the message
function sendEmail(subject, message) {
  console.log("Enter publishMessage");  

	var params = {
	  Message: message, 
	  Subject: subject,
	  TopicArn: topicArn
	};
	sns.publish(params, function(err, data) {
	  console.log("SNS callback...");  
	  if (err) {
	    console.log(err, err.stack); // an error occurred
	    throw err;
	  } else {
	    console.log(data);           // successful response
	  }
	});  
  
}


// Compare the new image and the old image objects
// If an account's status is SUSPENDED in the new image and the status is ACTIVE in the old image, the account is newly closed
// Params:
//   event: event object that has the new image and old image from DynamoDB Streams
function getClosedAccounts(event) {
	var accounts = [];

	for (const record of event.Records) {
		console.log(record.eventID);
		console.log(record.eventName);
		console.log('DynamoDB Record: %j', record.dynamodb);

		if (!record.dynamodb.NewImage || !record.dynamodb.OldImage) {
			console.log("Cannot compare. Missing either old image or new image.")
			continue;
		}

		console.log('New DynamoDB Record: %j', record.dynamodb.NewImage.Accounts.S);
		console.log('Old DynamoDB Record: %j', record.dynamodb.OldImage.Accounts.S);


		var newAccountStatus = JSON.parse(record.dynamodb.NewImage.Accounts.S);
		var oldAccountStatus = JSON.parse(record.dynamodb.OldImage.Accounts.S);
		 
		 
		for (const newStatus of newAccountStatus.Accounts) {
			if (newStatus.Status == "SUSPENDED") {
				for (const oldStatus of oldAccountStatus.Accounts) {
					if (newStatus.Id == oldStatus.Id && oldStatus.Status != "SUSPENDED") {
						console.log("Status changed");
						accounts.push(newStatus.Id);
					}
				}
			}
		}
	}

	return accounts;
}

exports.handler = (event, context, callback) => {
  console.log('entry handler');
  
  // Get the account ID from the Lambda function ARN
  var tempStr = context.invokedFunctionArn.substring(0,context.invokedFunctionArn.indexOf(":function:"+context.functionName));
  const destinationAccount = tempStr.substring(tempStr.lastIndexOf(":")+1);
  
  // Get closed accounts by comparing the new image and the old image in the event object
  var closedAccounts = getClosedAccounts(event);
  
  var sourceAccount;
  
  const date = new Date();
  // 12 months before the current date in the 'YYYY-MM-DD' format
  const startDate = date.getFullYear()-1 + "-" + formatToTwoDigits(date.getMonth()+1) + "-" + formatToTwoDigits(date.getDate());
  // Current date in the 'YYYY-MM-DD' format
  const endDate = date.getFullYear() + "-" +  formatToTwoDigits(date.getMonth()+1) + "-" + formatToTwoDigits(date.getDate());
 
  // Loop through the closed accounts and process them
  for (sourceAccount of closedAccounts) {
    console.log("account is " + sourceAccount);
    processReservationUtilization(startDate,endDate,sourceAccount,destinationAccount);
    processSavingsPlansUtilization(startDate,endDate,sourceAccount,destinationAccount);
  }  
}
